#include <stdlib.h>
#include "tund.h"
#include <openssl/blowfish.h>
#include <openssl/idea.h>
#include <openssl/rc5.h>


extern int debug;
extern Seq *stun;

static unsigned char ivec_ini[8]={"kjsLd5Gh"};
static unsigned char iv_buf[8];


static void *must_malloc(int size)
 {
  void *ret;

  ret=malloc(size);
  if(!ret)Error("Can not allocate %d bytes in must_malloc",size);
  return ret;
 }



void set_cipher_key(int tun_ind,int remote)
 {
  Tunnel *tun;
  BF_KEY *bf_key;
  IDEA_KEY_SCHEDULE *idea_key1,*idea_key2;
  RC5_32_KEY *rc5_key;
  void *key;
  unsigned char *data;
  int cipher;

  tun=(Tunnel*)stun->buf[tun_ind];

  cipher=(tun->flags)>>16;

  if(remote) 
   {
    if(tun->remoteCipherKey) free(tun->remoteCipherKey);
    tun->remoteCipherKey=NULL;
    data=tun->remote_key;
   }
  else 
   {
    if(tun->localCipherKey) free(tun->localCipherKey);
    tun->localCipherKey=NULL;
    data=tun->local_key;
   }

  switch(cipher)
   {
    case TUN_CIPHER_BF:	bf_key=must_malloc(sizeof(BF_KEY));
			BF_set_key(bf_key,16,data);
			key=bf_key;
			break;
    case TUN_CIPHER_IDEA: idea_key1=must_malloc(sizeof(IDEA_KEY_SCHEDULE));
			idea_set_encrypt_key(data,idea_key1);
			if(remote)
			 {
			  idea_key2=must_malloc(sizeof(IDEA_KEY_SCHEDULE));
			  idea_set_decrypt_key(idea_key1,idea_key2);
			  free(idea_key1);
			  idea_key1=idea_key2;
			 }
			key=idea_key1;
			break;
    case TUN_CIPHER_RC5:rc5_key=must_malloc(sizeof(RC5_32_KEY));
			RC5_32_set_key(rc5_key,16,data,RC5_8_ROUNDS);
			key=rc5_key;
			break;
    default:		Error("Unknown cipher %d",cipher);
   }

  if(remote) 
   tun->remoteCipherKey=key;
  else
   tun->localCipherKey=key;

#ifdef DEBUG
   if(debug)Log("New %s key for cipher %d for tunnel [%s]",
                 remote?"remote":"local",cipher,tun->label);
#endif /*DEBUG*/
 }




int cipher_encrypt(unsigned char *in,unsigned char *out,int len,int tun_ind)
 {
  Tunnel *tun=(Tunnel*)stun->buf[tun_ind];
  int cipher;
  BF_KEY *bf_key;
  IDEA_KEY_SCHEDULE *idea_key;
  RC5_32_KEY *rc5_key;
  int length;

  length=((len/8)+1)*8;

  cipher=(tun->flags)>>16;
  switch(cipher)
   {
    case TUN_CIPHER_BF:	bf_key=tun->localCipherKey;
			bcopy(ivec_ini,iv_buf,8);
			BF_cbc_encrypt(in,out,length,bf_key,iv_buf,BF_ENCRYPT);
			return length;
    case TUN_CIPHER_IDEA:idea_key=tun->localCipherKey;
			bcopy(ivec_ini,iv_buf,8);
			idea_cbc_encrypt(in,out,length,idea_key,
                                         iv_buf,IDEA_ENCRYPT);
			return length;
    case TUN_CIPHER_RC5:rc5_key=tun->localCipherKey;
			bcopy(ivec_ini,iv_buf,8);
			RC5_32_cbc_encrypt(in,out,length,rc5_key,
                                          iv_buf,RC5_ENCRYPT);
			return length;
    default:		Error("Unknown cipher %d",cipher);
   }
 }





void cipher_decrypt(unsigned char *in,unsigned char *out,int len,int tun_ind)
 {
  Tunnel *tun=(Tunnel*)stun->buf[tun_ind];
  int cipher;
  BF_KEY *bf_key;
  IDEA_KEY_SCHEDULE *idea_key;
  RC5_32_KEY *rc5_key;
  int length;

  length=len;

  cipher=(tun->flags)>>16;
  switch(cipher)
   {
    case TUN_CIPHER_BF:	bf_key=tun->remoteCipherKey;
			bcopy(ivec_ini,iv_buf,8);
			BF_cbc_encrypt(in,out,length,bf_key,iv_buf,BF_DECRYPT);
			break;
    case TUN_CIPHER_IDEA:idea_key=tun->remoteCipherKey;
			bcopy(ivec_ini,iv_buf,8);
			idea_cbc_encrypt(in,out,length,idea_key,
                                         iv_buf,IDEA_DECRYPT);
			break;
    case TUN_CIPHER_RC5:rc5_key=tun->remoteCipherKey;
			bcopy(ivec_ini,iv_buf,8);
			RC5_32_cbc_encrypt(in,out,length,rc5_key,
                                          iv_buf,RC5_DECRYPT);
			break;
    default:		Error("Unknown cipher %d",cipher);
   }
 }


