#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include "tund.h"


extern int debug;
extern Seq *stun;
extern int control_port;
extern int socket_control;

unsigned char packetBuf[PACKET_MAX_SIZE];

static void input_packet(int tun_ind,int s,int len);
static void input_control(int tun_ind,int s,int len);




static int ks_recvfrom(int s,void *buf,size_t len,int flags,
                       struct sockaddr *from,int *fromlen)
 {
  int ret;

  ret=recvfrom(s,buf,len,flags,from,fromlen);

#ifdef DEBUG
/*  if(debug) Log("Recvfrom packet on socket %d len=%d",s,ret);*/
#endif /*DEBUG*/
  if(ret<0)
   Error("Error in ks_recvfrom(s=%d) while recvfrom: %s",s,strerror(errno));

  return ret;
 }



void ks_sendto(int tun_ind,int s,void *msg,size_t len,int flags,
               struct sockaddr *to,int tolen)
 {
  int ret;

  ret=sendto(s,msg,len,flags,to,tolen);
  if(ret<0)
   {
    if(errno==EINVAL)
     {
      Log("Dropped illegal IP packet (t=%d,so=%d,l=%d)",tun_ind,s,len);
      return;
     }
    else
     Error("Error %d while sendto(t=%d,s=%d): %s",
            errno,tun_ind,s,strerror(errno));
   }
  if(ret!=len)
     Error("Error while sendto(t=%d,s=%d): %d!=%d",tun_ind,s,ret,len);

#ifdef DEBUG
/*  if(debug) Log("Sended %d bytes to socket %d",ret,s);*/
#endif /*DEBUG*/
 }




void read_packet(int tun_ind,int s)
 {
  struct sockaddr_in from;
  struct sockaddr_in saddr;
  int fromlen;
  Tunnel *tun;
  int len,ret;
  PacketHead *phead;

  fromlen=sizeof(from);

  len=ks_recvfrom(s,packetBuf,PACKET_MAX_SIZE,0,
                  (struct sockaddr*)&from,&fromlen);

  phead=(PacketHead*)packetBuf;

  if(len<sizeof(PacketHead)+1 || phead->ver!=PPF_PROTO_VER)
   {
    Log("Dropped packet (illegal packet) l=%d from %s:%d lh4=0x%x f=0x%x v=%d",
        len,addr2ascii(AF_INET,&(from.sin_addr),sizeof(struct in_addr),NULL),
        (int)ntohs(from.sin_port),phead->tunAddr,phead->flags,phead->ver);
    return;
   }

  tun_ind=find_tunnel_from(&from,phead->tunAddr);
  if(tun_ind<0)
   {
    Log("Dropped packet (unknown tunnel) l=%d from %s:%d lh4=0x%x f=0x%x v=%d",
        len,addr2ascii(AF_INET,&(from.sin_addr),sizeof(struct in_addr),NULL),
        (int)ntohs(from.sin_port),phead->tunAddr,phead->flags,phead->ver);
    return;
   }
   
  switch(phead->flags & 0x0003)
   {
    case PPF_DATA: input_packet(tun_ind,s,len);
		   break;
    case PPF_CNT:  input_control(tun_ind,s,len);
                   break;
    default:
        Log("Dropped packet len=%d from %s:%d lh4=0x%x flags=0x%x",
        len,addr2ascii(AF_INET,&(from.sin_addr),sizeof(struct in_addr),NULL),
                       (int)ntohs(from.sin_port),phead->tunAddr,phead->flags);
      return;

   }
 }




static void input_packet(int tun_ind,int s,int buf_len)
 {
  struct sockaddr_in saddr;
  int fromlen;
  Tunnel *tun;
  DP *data;
  EDP *en_data;
  int body_len;
  unsigned char *body;

  tun=(Tunnel*)stun->buf[tun_ind];

  if(tun->flags & (TUNF_ENC|TUNF_HAS_RPWD))
   {
    en_data=(EDP*)packetBuf;
    data_decrypt(packetBuf+sizeof(PacketHead),buf_len-sizeof(PacketHead),
                 tun_ind);
    body=en_data->body;
    body_len=en_data->body_len;
   }
  else
   {
    data=(DP*)packetBuf;
    body_len=buf_len-sizeof(PacketHead);
    body=data->body;
   }

#ifdef DEBUG
  if(debug) Log("Received %d bytes from tunnel [%s]",body_len,tun->label);
#endif /*DEBUG*/

  if(tun->state & TUNS_PRC)
   {

    ks_sendto(tun_ind,tun->socketDivert,body,body_len,0,
              (struct sockaddr*)&(tun->l_addr),sizeof(struct sockaddr_in));
   }
#ifdef DEBUG
  else
   if(debug) Log("Dropped %d bytes on tunnel [%s]",body_len,tun->label);
#endif /*DEBUG*/
 }




static void input_control(int tun_ind,int s,int buf_len)
 {
  int ret;
  Tunnel *tun;

  tun=(Tunnel*)stun->buf[tun_ind];

#ifdef DEBUG
  if(debug) Log("Received %d bytes from control of tunnel [%s]",
                 buf_len,tun->label);
#endif /*DEBUG*/

  proceed_cp(tun_ind,buf_len);
 }




void input_divert(int tun_ind,int s)
 {
  int buf_len;
  int len;
  Tunnel *tun;
  struct sockaddr from;
  int fromlen=sizeof(from);
  DP *data;
  EDP *en_data;

  tun=(Tunnel*)stun->buf[tun_ind];

#ifdef DEBUG
  if(debug) Log("Input packet on divert socket %d",s);
#endif /*DEBUG*/

  if(tun->flags & TUNF_ENC)
   {
    en_data=(EDP*)packetBuf;
    len=ks_recvfrom(s,en_data->body,BUF_SIZE,0,&from,&fromlen);
    en_data->h.tunAddr=tun->tunAddr;
    en_data->h.flags=PPF_DATA;
    en_data->h.ver=PPF_PROTO_VER;
    en_data->rn=get_rand();
    en_data->body_len=(unsigned short)len;
    buf_len=data_encrypt(packetBuf+sizeof(PacketHead),len+4,tun_ind);
    buf_len+=sizeof(PacketHead);
   }
  else
   {
    data=(DP*)packetBuf;
    len=ks_recvfrom(s,data->body,BUF_SIZE,0,&from,&fromlen);
    buf_len=len+sizeof(PacketHead);
    data->h.tunAddr=tun->tunAddr;
    data->h.flags=PPF_DATA;
    data->h.ver=PPF_PROTO_VER;
   }

  if(tun->state & TUNS_PRC)
   {
    ks_sendto(tun_ind,socket_control,packetBuf,buf_len,0,
              (struct sockaddr*)(&(tun->to)),sizeof(struct sockaddr_in));
   }
#ifdef DEBUG
   else
    if(debug) Log("Dropped %d bytes on tunnel [%s]",len,tun->label);
#endif /*DEBUG*/
 }

