/******************************************************************************
* tund v0.20
* Simple IP over UDP tunnel assistant.
* Copyright Sergey Kosyakov ks@itp.ac.ru 1999
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <netdb.h>
#include <poll.h>
#include <time.h>
#include <errno.h>
#include "tund.h"

static void run_loop(int sc);
static void do_cycle(struct pollfd *fds,int nfd,CBF *cb);

extern int debug;
int control_port;
int socket_control;
extern Seq *stun;


static void proceed_tunnel(int tun_ind)
 {
  int sd;
  struct sockaddr_in saddr;
  int res;
  int i;
  Tunnel *tun,*t;

  tun=(Tunnel*)stun->buf[tun_ind];

  for(i=0;i<tun_ind;i++)
   {
    t=(Tunnel*)stun->buf[i];
    if(t->dp==tun->dp)
     Error("Illegal configuration: tunnels should have different divert sockets");
    if(t->tunAddr==tun->tunAddr)
     Error("Illegal configuration: tunnels should have different labels");
   }

  bzero(&saddr,sizeof(saddr));
  saddr.sin_port=htons((unsigned short)tun->dp);
  saddr.sin_family=AF_INET;
  sd=socket(PF_INET,SOCK_RAW,IPPROTO_DIVERT);
  res=bind(sd,(struct sockaddr*)&saddr,sizeof(saddr));
  if(res)
   Error("Can not bind to divert port %d\n",tun->dp);

  /*Set to non-blocking*/
  fcntl(sd,F_SETFL,FNDELAY);

  tun->socketDivert=sd;

#ifdef DEBUG
  if(debug) Log("Divert port %d is binded to socket %d for t=%d",
                 tun->dp,tun->socketDivert,tun_ind);
#endif DEBUG
 }



void do_it()
 {
  int i;
  int sp,sc;
  struct sockaddr_in saddr;
  int res;

  sp=control_port;
  bzero(&saddr,sizeof(saddr));
  saddr.sin_port=htons((unsigned short)sp);
  saddr.sin_family=AF_INET;
  saddr.sin_addr.s_addr=htonl(INADDR_ANY);
  sc=socket(PF_INET,SOCK_DGRAM,0);
  res=bind(sc,(struct sockaddr*)&saddr,sizeof(saddr));
  if(res)
   Error("Can not bind to control port %d\n",sp);
  /*Non-blocking*/
  fcntl(sc,F_SETFL,FNDELAY);
  
  /*Set up tunnel(s) divert sockets*/
  for(i=0;i<stun->length;i++)
   proceed_tunnel(i);

  Log("sc=%d",sc);

  run_loop(sc);
 }



static void run_loop(int sc)
 {
  struct pollfd *fds;
  CBF *cb;
  int nfd;
  int i;
  Tunnel *tun;

  socket_control=sc;

  nfd=1+stun->length;

  fds=(struct pollfd*)malloc(nfd*sizeof(struct pollfd));
  if(!fds)
   Error("Can not allocate memory for pollfd");
  bzero(fds,nfd*sizeof(struct pollfd));

  cb=(CBF*)malloc(nfd*sizeof(CBF));
  if(!cb)
   Error("Can not allocate memory for callbacks");
  bzero(cb,nfd*sizeof(CBF));

  for(i=0;i<stun->length;i++)
   {
    tun=(Tunnel*)stun->buf[i];
    fds[i].fd=tun->socketDivert;
    fds[i].events=POLLRDNORM;
    cb[i].tun_ind=i;
    cb[i].f=input_divert;
    if(tun->flags & TUNF_ENC)
    /* Initiate encryption for that tunnel */
      start_secur_tunnel(i);
    else
      tun->state=TUNS_PRC;
   }

  fds[stun->length].fd=sc;
  fds[stun->length].events=POLLRDNORM;
  cb[stun->length].f=read_packet;
  cb[stun->length].tun_ind=-1; /*Does have no any means*/

  do_cycle(fds,nfd,cb);
 }



static void do_cycle(struct pollfd *fds,int nfd,CBF *cb)
 {
  int ret;
  int i;
  time_t l_time;
  Tunnel *tun;

  do /*Main loop*/
   {
    ret=poll(fds,nfd,15000);
    if(ret<0)
     {
      if(errno==EINTR) continue;
      Error("Poll returns error %d",ret);
     }
#ifdef DEBUG
    if(debug) Log("Poll returns %d",ret); 
#endif /*DEBUG*/

    for(i=0;i<nfd;i++)
     {
      if(fds[i].revents)
       {
#ifdef DEBUG
        if(debug) Log("Event 0x%x on fd %d",fds[i].revents,fds[i].fd);
#endif /*DEBUG*/
        cb[i].f(cb[i].tun_ind,fds[i].fd);
       }
     }
   } while(1); /*End of main loop*/
 }


